

from string import Formatter
from typing import List


BUILTIN_ARGUMENTS = {"interface_name", "service_id", "signal_name", "property_name", "method_name", "client_id"}

def is_valid_topic_template(template: str, additional_arguments: List[str]) -> bool:
    """Check if a topic template is valid.
    
    A template is valid if:
    1. All arguments are from BUILTIN_ARGUMENTS or additional_arguments
    2. No argument appears more than once in the template
    
    Args:
        template: The template string to validate.
        additional_arguments: List of additional allowed argument names.
        
    Returns:
        True if the template is valid, False otherwise.
    """
    formatter = Formatter()
    allowed_arguments = BUILTIN_ARGUMENTS | set(additional_arguments)
    seen_arguments = set()
    
    for _, field_name, _, _ in formatter.parse(template):
        if field_name is not None:
            # Check if argument already appeared
            if field_name in seen_arguments:
                return False
            
            # Check if argument is allowed
            if field_name not in allowed_arguments:
                return False
            
            seen_arguments.add(field_name)
    
    return True


def topic_template_fill_in(template: str, **kwargs) -> str:
    """Fill in a topic template with the provided keyword arguments.
    
    If only some template arguments are provided, they will be replaced
    while others remain as template placeholders.
    """
    formatter = Formatter()
    result = []
    for literal_text, field_name, format_spec, conversion in formatter.parse(template):
        result.append(literal_text)
        if field_name is not None:
            if field_name in kwargs:
                value = kwargs[field_name]
                if conversion:
                    value = formatter.convert_field(value, conversion)
                if format_spec:
                    value = format(value, format_spec)
                result.append(str(value))
            else:
                # Keep the placeholder for unprovided arguments
                placeholder = '{' + field_name
                if conversion:
                    placeholder += '!' + conversion
                if format_spec:
                    placeholder += ':' + format_spec
                placeholder += '}'
                result.append(placeholder)
    return ''.join(result)


def get_topic_arguments(template: str) -> list[str]:
    """Get a list of argument names in the order they appear in the template.
    
    Args:
        template: The template string to extract arguments from.
        
    Returns:
        A list of argument names in the order they appear.
    """
    formatter = Formatter()
    arguments = []
    for _, field_name, _, _ in formatter.parse(template):
        if field_name is not None and field_name not in arguments:
            arguments.append(field_name)
    return arguments


def get_argument_position(template: str, argument: str) -> int | None:
    """Get the position of an argument in the topic when split by '/'.
    
    Args:
        template: The template string to search.
        argument: The argument name to find.
        
    Returns:
        The index of the segment containing the argument when the topic is split by '/',
        or None if the argument is not in the template.
    """
    formatter = Formatter()
    segment_index = 0
    
    for literal_text, field_name, _, _ in formatter.parse(template):
        # Count slashes in literal text before this field
        if literal_text:
            segment_index += literal_text.count('/')
        
        if field_name is not None and field_name == argument:
            return segment_index
    
    return None




from stingeripc.topic_util import (
    topic_template_fill_in,
    get_topic_arguments,
    is_valid_topic_template,
    get_argument_position,
)

import unittest


class TestTopicTemplateFillIn(unittest.TestCase):
    def test_full_replacement(self):
        """Test when all template arguments are provided."""
        template = "hello/{name}/world/{place}"
        result = topic_template_fill_in(template, name="Alice", place="Paris")
        self.assertEqual(result, "hello/Alice/world/Paris")

    def test_partial_replacement(self):
        """Test when only some template arguments are provided."""
        template = "hello/{name}/world/{place}"
        result = topic_template_fill_in(template, name="Alice")
        self.assertEqual(result, "hello/Alice/world/{place}")

    def test_no_replacement(self):
        """Test when no template arguments are provided."""
        template = "hello/{name}/world/{place}"
        result = topic_template_fill_in(template)
        self.assertEqual(result, "hello/{name}/world/{place}")

    def test_empty_template(self):
        """Test with an empty template string."""
        result = topic_template_fill_in("")
        self.assertEqual(result, "")

    def test_no_placeholders(self):
        """Test with a template that has no placeholders."""
        template = "hello/world"
        result = topic_template_fill_in(template, name="Alice")
        self.assertEqual(result, "hello/world")

    def test_single_placeholder(self):
        """Test with a single placeholder."""
        template = "{name}"
        result = topic_template_fill_in(template, name="Bob")
        self.assertEqual(result, "Bob")

    def test_multiple_same_placeholder(self):
        """Test with the same placeholder appearing multiple times."""
        template = "{name}/test/{name}"
        result = topic_template_fill_in(template, name="Alice")
        self.assertEqual(result, "Alice/test/Alice")

    def test_partial_with_multiple_same_placeholder(self):
        """Test partial replacement with repeated placeholders."""
        template = "{name}/test/{name}/{place}"
        result = topic_template_fill_in(template, name="Alice")
        self.assertEqual(result, "Alice/test/Alice/{place}")

    def test_format_spec_provided(self):
        """Test with format specification when value is provided."""
        template = "{value:.2f}"
        result = topic_template_fill_in(template, value=3.14159)
        self.assertEqual(result, "3.14")

    def test_format_spec_not_provided(self):
        """Test that format spec is preserved when value is not provided."""
        template = "{value:.2f}"
        result = topic_template_fill_in(template)
        self.assertEqual(result, "{value:.2f}")

    def test_conversion_provided(self):
        """Test with conversion (!s, !r, !a) when value is provided."""
        template = "{name!r}"
        result = topic_template_fill_in(template, name="Alice")
        self.assertEqual(result, "'Alice'")

    def test_conversion_not_provided(self):
        """Test that conversion is preserved when value is not provided."""
        template = "{name!r}"
        result = topic_template_fill_in(template)
        self.assertEqual(result, "{name!r}")

    def test_conversion_and_format_spec(self):
        """Test with both conversion and format spec."""
        template = "{value!s:>10}"
        result = topic_template_fill_in(template, value=42)
        self.assertEqual(result, "        42")

    def test_conversion_and_format_spec_not_provided(self):
        """Test that both conversion and format spec are preserved."""
        template = "{value!s:>10}"
        result = topic_template_fill_in(template)
        self.assertEqual(result, "{value!s:>10}")

    def test_mixed_provided_and_not_provided(self):
        """Test complex case with multiple placeholders, some provided, some not."""
        template = "{service}/{version}/api/{endpoint}/{id}"
        result = topic_template_fill_in(template, service="users", endpoint="profile")
        self.assertEqual(result, "users/{version}/api/profile/{id}")

    def test_numeric_values(self):
        """Test with numeric values."""
        template = "device/{device_id}/sensor/{sensor_id}"
        result = topic_template_fill_in(template, device_id=123, sensor_id=456)
        self.assertEqual(result, "device/123/sensor/456")

    def test_extra_kwargs_ignored(self):
        """Test that extra kwargs that don't match placeholders are ignored."""
        template = "{name}"
        result = topic_template_fill_in(template, name="Alice", extra="ignored")
        self.assertEqual(result, "Alice")


class TestGetTopicArguments(unittest.TestCase):
    def test_single_argument(self):
        """Test extracting a single argument."""
        template = "{name}"
        result = get_topic_arguments(template)
        self.assertEqual(result, ["name"])

    def test_multiple_arguments(self):
        """Test extracting multiple arguments in order."""
        template = "{service}/{version}/api/{endpoint}/{id}"
        result = get_topic_arguments(template)
        self.assertEqual(result, ["service", "version", "endpoint", "id"])

    def test_no_arguments(self):
        """Test template with no arguments."""
        template = "hello/world"
        result = get_topic_arguments(template)
        self.assertEqual(result, [])

    def test_empty_template(self):
        """Test with empty template."""
        template = ""
        result = get_topic_arguments(template)
        self.assertEqual(result, [])

    def test_repeated_arguments(self):
        """Test that repeated arguments only appear once."""
        template = "{name}/test/{name}/{place}"
        result = get_topic_arguments(template)
        self.assertEqual(result, ["name", "place"])

    def test_arguments_with_format_spec(self):
        """Test that arguments with format specs are extracted correctly."""
        template = "{value:.2f}/{count:>5}"
        result = get_topic_arguments(template)
        self.assertEqual(result, ["value", "count"])

    def test_arguments_with_conversion(self):
        """Test that arguments with conversions are extracted correctly."""
        template = "{name!r}/{value!s}"
        result = get_topic_arguments(template)
        self.assertEqual(result, ["name", "value"])

    def test_complex_template(self):
        """Test with a complex template."""
        template = "device/{device_id}/sensor/{sensor_id}/data/{timestamp}"
        result = get_topic_arguments(template)
        self.assertEqual(result, ["device_id", "sensor_id", "timestamp"])


class TestIsValidTopicTemplate(unittest.TestCase):
    def test_valid_builtin_arguments(self):
        """Test template with only builtin arguments is valid."""
        template = "{interface_name}/signal/{signal_name}"
        result = is_valid_topic_template(template, [])
        self.assertTrue(result)

    def test_valid_additional_arguments(self):
        """Test template with additional arguments is valid."""
        template = "{interface_name}/custom/{custom_arg}"
        result = is_valid_topic_template(template, ["custom_arg"])
        self.assertTrue(result)

    def test_valid_mixed_arguments(self):
        """Test template with both builtin and additional arguments."""
        template = "{interface_name}/{service_id}/data/{sensor_id}"
        result = is_valid_topic_template(template, ["sensor_id"])
        self.assertTrue(result)

    def test_invalid_unknown_argument(self):
        """Test template with unknown argument is invalid."""
        template = "{interface_name}/test/{unknown_arg}"
        result = is_valid_topic_template(template, [])
        self.assertFalse(result)

    def test_invalid_duplicate_argument(self):
        """Test template with duplicate argument is invalid."""
        template = "{interface_name}/test/{interface_name}"
        result = is_valid_topic_template(template, [])
        self.assertFalse(result)

    def test_invalid_duplicate_additional_argument(self):
        """Test template with duplicate additional argument is invalid."""
        template = "{custom_arg}/test/{custom_arg}"
        result = is_valid_topic_template(template, ["custom_arg"])
        self.assertFalse(result)

    def test_valid_empty_template(self):
        """Test empty template is valid."""
        template = ""
        result = is_valid_topic_template(template, [])
        self.assertTrue(result)

    def test_valid_no_arguments(self):
        """Test template with no arguments is valid."""
        template = "static/topic/path"
        result = is_valid_topic_template(template, [])
        self.assertTrue(result)

    def test_valid_all_builtin_arguments(self):
        """Test template using all builtin arguments."""
        template = "{interface_name}/{service_id}/{signal_name}/{property_name}/{method_name}/{client_id}"
        result = is_valid_topic_template(template, [])
        self.assertTrue(result)

    def test_invalid_partial_match(self):
        """Test that partial matching doesn't work - argument must be exact."""
        template = "{interface_name_extra}"
        result = is_valid_topic_template(template, [])
        self.assertFalse(result)

    def test_valid_with_format_spec(self):
        """Test that format specs don't affect validation."""
        template = "{interface_name}/{service_id:.5}"
        result = is_valid_topic_template(template, [])
        self.assertTrue(result)

    def test_valid_with_conversion(self):
        """Test that conversions don't affect validation."""
        template = "{interface_name!r}/{signal_name!s}"
        result = is_valid_topic_template(template, [])
        self.assertTrue(result)


class TestGetArgumentPosition(unittest.TestCase):
    def test_property_name_at_index_2(self):
        """Test that property_name is at index 2 in the template."""
        template = "{service_id}/properties/{property_name}/update"
        result = get_argument_position(template, "property_name")
        self.assertEqual(result, 2)

    def test_nonexistent_argument_returns_none(self):
        """Test that a nonexistent argument returns None."""
        template = "{service_id}/properties/{property_name}/update"
        result = get_argument_position(template, "property_name_2")
        self.assertIsNone(result)

    def test_first_argument_at_index_0(self):
        """Test argument at the beginning of template."""
        template = "{interface_name}/signal/{signal_name}"
        result = get_argument_position(template, "interface_name")
        self.assertEqual(result, 0)

    def test_last_argument_position(self):
        """Test argument at the end of template."""
        template = "{interface_name}/signal/{signal_name}"
        result = get_argument_position(template, "signal_name")
        self.assertEqual(result, 2)

    def test_single_argument_template(self):
        """Test template with only one argument."""
        template = "{service_id}"
        result = get_argument_position(template, "service_id")
        self.assertEqual(result, 0)

    def test_no_slashes_template(self):
        """Test template without slashes."""
        template = "prefix{arg}suffix"
        result = get_argument_position(template, "arg")
        self.assertEqual(result, 0)

    def test_multiple_slashes_before_argument(self):
        """Test argument after multiple path segments."""
        template = "a/b/c/d/{target}/e/f"
        result = get_argument_position(template, "target")
        self.assertEqual(result, 4)

    def test_empty_template(self):
        """Test with empty template."""
        template = ""
        result = get_argument_position(template, "anything")
        self.assertIsNone(result)

    def test_template_without_arguments(self):
        """Test template with no placeholders."""
        template = "static/path/to/resource"
        result = get_argument_position(template, "anything")
        self.assertIsNone(result)

    def test_complex_template_multiple_arguments(self):
        """Test complex template with multiple arguments."""
        template = "{interface_name}/{service_id}/method/{method_name}/request"
        self.assertEqual(get_argument_position(template, "interface_name"), 0)
        self.assertEqual(get_argument_position(template, "service_id"), 1)
        self.assertEqual(get_argument_position(template, "method_name"), 3)

    def test_argument_with_format_spec(self):
        """Test that format specs don't affect position."""
        template = "{interface_name}/{value:.2f}/data"
        result = get_argument_position(template, "value")
        self.assertEqual(result, 1)

    def test_argument_with_conversion(self):
        """Test that conversions don't affect position."""
        template = "{interface_name}/{name!r}/info"
        result = get_argument_position(template, "name")
        self.assertEqual(result, 1)

    def test_consecutive_arguments_same_segment(self):
        """Test multiple arguments in the same segment."""
        template = "data/{interface_name}_{service_id}/resource"
        # Both should be in segment 1
        self.assertEqual(get_argument_position(template, "interface_name"), 1)
        self.assertEqual(get_argument_position(template, "service_id"), 1)


if __name__ == "__main__":
    unittest.main()



from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("stinger-ipc")
except PackageNotFoundError:
    __version__ = "unknown"

from .interface import StingerInterface
from .asyncapi import StingerToAsyncApi

__all__ = ["StingerInterface", "StingerToAsyncApi", "__version__"]



"""Configuration models for Stinger IPC."""

from pydantic import BaseModel, Field, ConfigDict, field_validator, model_validator
from pathlib import Path
from typing import List, Optional
import tomllib
from .topic_util import get_argument_position
import re

class ServerConfig(BaseModel):
    """Configuration options for generating server/provider code."""
    enabled: bool = Field(default=True, description="Whether to generate server code")


class ClientConfig(BaseModel):
    """Configuration options for generating client code."""
    enabled: bool = Field(default=True, description="Whether to generate client code")

class PropertyConfig(BaseModel):
    """Configuration options for properties."""
    stinger_owned_values: bool = Field(default=True, description="Whether the stinger-server owns property values")

class PythonConfig(BaseModel):
    """Python-specific configuration options."""
    python37: bool = Field(default=False, description="Generate Python 3.7 compatible code")

class CppConfig(BaseModel):
    """C++-specific configuration options."""
    namespace: List[str] = Field(default_factory=lambda: ["stinger", "gen"], description="List of nested namespaces for generated C++ code.  The interface name is appended to this as the innermost namespace.")

class TopicConfig(BaseModel):
    """ Configuration for which topic schemas to use """
    params: List[str] = Field(default_factory=list, description="List of parameters to include in topic templates")
    signals: str = Field(default="{interface_name}/{service_id}/signal/{signal_name}", description="Topic template for signals")
    property_values: str = Field(default="{interface_name}/{service_id}/property/{property_name}/value", description="Topic template for property values")
    property_updates: str = Field(default="{interface_name}/{service_id}/property/{property_name}/update", description="Topic template for property updates")
    property_update_responses: str = Field(default="client/{client_id}/{interface_name}/property/{property_name}/update/response", description="Topic template for property update responses")
    method_requests: str = Field(default="{interface_name}/{service_id}/method/{method_name}/request", description="Topic template for method requests")
    method_responses: str = Field(default="client/{client_id}/{interface_name}/method/{method_name}/response", description="Topic template for method responses")
    interface_discovery: str = Field(default="{interface_name}/{service_id}/interface", description="Topic template for interface discovery")

    @field_validator('property_values')
    @classmethod
    def validate_property_values(cls, v: str) -> str:
        if get_argument_position(v, 'property_name') is None:
            raise ValueError('"property_values" topic template must contain {property_name} placeholder')
        if get_argument_position(v, 'service_id') is None:
            raise ValueError('"property_values" topic template must contain {service_id} placeholder')
        return v
    
    @field_validator('property_update_responses')
    @classmethod
    def validate_property_update_responses(cls, v: str) -> str:
        if get_argument_position(v, 'property_name') is None:
            raise ValueError('"property_update_responses" topic template must contain {property_name} placeholder')
        return v
    
    @field_validator('method_responses')
    @classmethod
    def validate_method_responses(cls, v: str) -> str:
        if get_argument_position(v, 'method_name') is None:
            raise ValueError('"method_responses" topic template must contain {method_name} placeholder')
        return v

    @field_validator('params')
    @classmethod
    def validate_params(cls, v: List[str]) -> List[str]:
        reserved_params = {'interface_name', 'service_id', 'signal_name', 'property_name', 'method_name', 'client_id', 'instance_id'}
        for param in v:
            if param in reserved_params:
                raise ValueError(f'Custom topic parameters cannot use reserved names: {param}')
        for param in v:
            if not re.match(r'^[a-zA-Z0-9]+(_[a-zA-Z0-9]+)*$', param):
                raise ValueError(f'Custom topic parameters must be alphanumeric with optional underscores, but no consecutive underscores: {param}')
        return v

    @model_validator(mode='after')
    def validate_topic_values(self) -> 'TopicConfig':
        if self.property_update_responses == self.method_responses:
            raise ValueError('"property_update_responses" and "method_responses" topic templates must be different to avoid routing conflicts')
        return self

class LanguagePluginConfig(BaseModel):
    """ Configuration for languages that are provided via plugins """
    name: str = Field(..., description="Name of the language plugin")
    model_config = ConfigDict(extra='allow')

class DiscoveryConfig(BaseModel):
    """Configuration options for service discovery."""
    advert_interval_seconds: int = Field(default=120, description="Interval in seconds for re-advertising the server's presence")

class StingerConfig(BaseModel):
    """Root configuration model for Stinger IPC code generation."""
    python: PythonConfig = Field(default_factory=PythonConfig, description="Python generation options")
    cpp: CppConfig = Field(default_factory=CppConfig, description="C++ generation options")
    properties: PropertyConfig = Field(default_factory=PropertyConfig, description="Property generation options")
    server: ServerConfig = Field(default_factory=ServerConfig, description="Server code generation options")
    client: ClientConfig = Field(default_factory=ClientConfig, description="Client code generation options")
    topics: TopicConfig = Field(default_factory=TopicConfig, description="Topic schema configuration")
    discovery: DiscoveryConfig = Field(default_factory=DiscoveryConfig, description="Service discovery configuration")
    language: dict[str, LanguagePluginConfig] = Field(default_factory=dict, description="Language plugin configurations")
    model_config = ConfigDict(strict=True)


def load_config(config_path: Path) -> StingerConfig:
    """
    Load and parse a TOML configuration file.
    
    Args:
        config_path: Path to the TOML configuration file.
        
    Returns:
        A StingerConfig instance with the parsed configuration.
        
    Raises:
        FileNotFoundError: If the config file doesn't exist.
        ValueError: If the TOML is invalid or doesn't match the schema.
    """
    if not config_path.exists():
        raise FileNotFoundError(f"Configuration file not found: {config_path}")
    
    with config_path.open(mode="rb") as f:
        config_dict = tomllib.load(f)
    
    return StingerConfig(**config_dict)



from typing import Optional, Any, Dict, IO, Union
import yaml
import yamlloader

from .components import StingerSpec, InvalidStingerStructure
from .config import StingerConfig

VERSIONS_SUPPORTED = ["0.0.7", "0.1.0"]


class StingerInterface(StingerSpec):

    def __init__(self, stinger: Dict[str, Any], config: StingerConfig):
        super().__init__(stinger["interface"], config)

    @classmethod
    def from_yaml(cls, yaml_input: Union[str, IO], config: StingerConfig) -> StingerSpec:
        yaml_obj = yaml.load(yaml_input, Loader=yamlloader.ordereddict.Loader)
        return cls.new_spec_from_stinger(yaml_obj, config)

    @classmethod
    def from_dict(cls, stinger_dict: Dict[str, Any], config: StingerConfig) -> StingerSpec:
        return cls.new_spec_from_stinger(stinger_dict, config)